from kopf.structs.patches import Patch


def test_addition_of_the_key():
    patch = Patch()
    patch['xyz'] = 123
    jsonpatch = patch.as_json_patch()
    assert jsonpatch == [
        {'op': 'replace', 'path': '/xyz', 'value': 123},
    ]


def test_removal_of_the_key():
    patch = Patch()
    patch['xyz'] = None
    jsonpatch = patch.as_json_patch()
    assert jsonpatch == [
        {'op': 'remove', 'path': '/xyz'},
    ]


def test_addition_of_the_subkey():
    patch = Patch()
    patch['xyz'] = {'abc': 123}
    jsonpatch = patch.as_json_patch()
    assert jsonpatch == [
        {'op': 'replace', 'path': '/xyz/abc', 'value': 123},
    ]


def test_removal_of_the_subkey():
    patch = Patch()
    patch['xyz'] = {'abc': None}
    jsonpatch = patch.as_json_patch()
    assert jsonpatch == [
        {'op': 'remove', 'path': '/xyz/abc'},
    ]



import asyncio
import dataclasses
import warnings
from unittest.mock import Mock

import pyngrok.conf
import pyngrok.ngrok
import pytest

from kopf.reactor.indexing import OperatorIndexers
from kopf.structs.references import Insights, Resource
from kopf.structs.reviews import CreateOptions, Request, RequestKind, RequestPayload, \
                                 RequestResource, UserInfo, WebhookFn
from kopf.toolkits.webhooks import WebhookServer


# TODO: LATER: Fix this issue some day later.
@pytest.fixture()
def no_serverside_resource_warnings():
    """
    Hide an irrelevant ResourceWarning on the server side:

    It happens when a client disconnects from the webhook server,
    and the server closes the transport for that client. The garbage
    collector calls ``__del__()`` on the SSL proto object, despite
    it is not close to the moment.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',
                                category=ResourceWarning,
                                module='asyncio.sslproto',
                                message='unclosed transport')
        yield


# TODO: LATER: Fix this issue after aiohttp 4.0.0 is used.
@pytest.fixture()
async def no_clientside_resource_warnings():
    """
    Hide an irrelevant ResourceWarning on the client side.

    https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
    """
    yield
    await asyncio.sleep(0.100)


@pytest.fixture()
async def no_sslproto_warnings(no_serverside_resource_warnings, no_clientside_resource_warnings):
    pass


# cert generation is somewhat slow (~1s)
@pytest.fixture(scope='module')
def certpkey():
    cert, pkey = WebhookServer.build_certificate(['localhost', '127.0.0.1'])
    return cert, pkey


@pytest.fixture()
def certfile(tmpdir, certpkey):
    path = tmpdir.join('cert.pem')
    path.write_binary(certpkey[0])
    return str(path)


@pytest.fixture()
def pkeyfile(tmpdir, certpkey):
    path = tmpdir.join('pkey.pem')
    path.write_binary(certpkey[1])
    return str(path)


@pytest.fixture()
def adm_request(resource):
    return Request(
        apiVersion='admission.k8s.io/v1',
        kind='AdmissionReview',
        request=RequestPayload(
            uid='uid1',
            kind=RequestKind(group=resource.group, version=resource.version, kind=resource.kind),
            resource=RequestResource(group=resource.group, version=resource.version, resource=resource.plural),
            requestKind=RequestKind(group=resource.group, version=resource.version, kind=resource.kind),
            requestResource=RequestResource(group=resource.group, version=resource.version, resource=resource.plural),
            userInfo=UserInfo(username='user1', uid='useruid1', groups=['group1']),
            name='name1',
            namespace='ns1',
            operation='CREATE',
            options=CreateOptions(apiVersion='meta.k8s.io/v1', kind='CreateOptions'),
            object={'spec': {'field': 'value'}},
            oldObject=None,
            dryRun=False,
        ))


@dataclasses.dataclass(frozen=True)
class Responder:
    fn: WebhookFn
    fut: asyncio.Future  # asyncio.Future[Response]


@pytest.fixture()
async def responder() -> Responder:
    fut = asyncio.Future()
    async def fn(*_, **__):
        return await fut
    return Responder(fn=fn, fut=fut)


@pytest.fixture()
async def insights(settings, resource):
    val_resource = Resource('admissionregistration.k8s.io', 'v1', 'validatingwebhookconfigurations')
    mut_resource = Resource('admissionregistration.k8s.io', 'v1', 'mutatingwebhookconfigurations')
    insights = Insights()
    insights.resources.add(resource)
    await insights.backbone.fill(resources=[val_resource, mut_resource])
    insights.ready_resources.set()
    return insights


@pytest.fixture()
def indices():
    indexers = OperatorIndexers()
    return indexers.indices


@dataclasses.dataclass(frozen=True, eq=False)
class K8sMocks:
    patch_obj: Mock
    create_obj: Mock
    post_event: Mock
    sleep_or_wait: Mock


@pytest.fixture(autouse=True)
def k8s_mocked(mocker):
    # We mock on the level of our own K8s API wrappers, not the K8s client.
    return K8sMocks(
        patch_obj=mocker.patch('kopf.clients.patching.patch_obj', return_value={}),
        create_obj=mocker.patch('kopf.clients.creating.create_obj', return_value={}),
        post_event=mocker.patch('kopf.clients.events.post_event'),
        sleep_or_wait=mocker.patch('kopf.structs.primitives.sleep_or_wait', return_value=None),
    )


@pytest.fixture(autouse=True)
def pyngrok_mock(mocker):
    mocker.patch.object(pyngrok.conf, 'get_default')
    mocker.patch.object(pyngrok.ngrok, 'set_auth_token')
    mocker.patch.object(pyngrok.ngrok, 'connect')
    mocker.patch.object(pyngrok.ngrok, 'disconnect')
    pyngrok.ngrok.connect.return_value.public_url = 'https://nowhere'
    return pyngrok



from typing import Optional, cast

from kopf.clients import auth, errors
from kopf.structs import bodies, references


@auth.reauthenticated_request
async def create_obj(
        *,
        resource: references.Resource,
        namespace: references.Namespace = None,
        name: Optional[str] = None,
        body: Optional[bodies.RawBody] = None,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> Optional[bodies.RawBody]:
    """
    Create a resource.
    """
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    body = body if body is not None else {}
    if namespace is not None:
        body.setdefault('metadata', {}).setdefault('namespace', namespace)
    if name is not None:
        body.setdefault('metadata', {}).setdefault('name', name)

    namespace = cast(references.Namespace, body.get('metadata', {}).get('namespace'))
    response = await context.session.post(
        url=resource.get_url(server=context.server, namespace=namespace),
        json=body,
    )
    created_body: bodies.RawBody = await errors.parse_response(response)
    return created_body



import asyncio
import ssl
import urllib.parse
from typing import Collection, Mapping, Optional, Set, Tuple

from kopf.clients import auth, errors
from kopf.structs import references


@auth.reauthenticated_request
async def read_sslcert(
        *,
        context: Optional[auth.APIContext] = None,
) -> Tuple[str, bytes]:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    parsed = urllib.parse.urlparse(context.server)
    host = parsed.hostname or ''  # NB: it cannot be None/empty in our case.
    port = parsed.port or 443
    loop = asyncio.get_running_loop()
    cert = await loop.run_in_executor(None, ssl.get_server_certificate, (host, port))
    return host, cert.encode('ascii')


@auth.reauthenticated_request
async def read_version(
        *,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> Mapping[str, str]:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    server = context.server.rstrip('/')
    url = f'{server}/version'
    rsp: Mapping[str, str] = await errors.parse_response(await context.session.get(url))
    return rsp


@auth.reauthenticated_request
async def scan_resources(
        *,
        groups: Optional[Collection[str]] = None,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> Collection[references.Resource]:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")
    coros = {
        _read_old_api(groups=groups, context=context),
        _read_new_apis(groups=groups, context=context),
    }
    resources: Set[references.Resource] = set()
    for coro in asyncio.as_completed(coros):
        resources.update(await coro)
    return resources


async def _read_old_api(
        *,
        groups: Optional[Collection[str]],
        context: auth.APIContext,
) -> Collection[references.Resource]:
    resources: Set[references.Resource] = set()
    if groups is None or '' in groups:
        server = context.server.rstrip('/')
        url = f'{server}/api'
        rsp = await errors.parse_response(await context.session.get(url))
        coros = {
            _read_version(
                url=f'{server}/api/{version_name}',
                group='',
                version=version_name,
                preferred=True,
                context=context)
            for version_name in rsp['versions']
        }
        for coro in asyncio.as_completed(coros):
            resources.update(await coro)
    return resources


async def _read_new_apis(
        *,
        groups: Optional[Collection[str]],
        context: auth.APIContext,
) -> Collection[references.Resource]:
    resources: Set[references.Resource] = set()
    if groups is None or set(groups or {}) - {''}:
        server = context.server.rstrip('/')
        url = f'{server}/apis'
        rsp = await errors.parse_response(await context.session.get(url))
        items = [d for d in rsp['groups'] if groups is None or d['name'] in groups]
        coros = {
            _read_version(
                url=f'{server}/apis/{group_dat["name"]}/{version["version"]}',
                group=group_dat['name'],
                version=version['version'],
                preferred=version['version'] == group_dat['preferredVersion']['version'],
                context=context)
            for group_dat in items
            for version in group_dat['versions']
        }
        for coro in asyncio.as_completed(coros):
            resources.update(await coro)
    return resources


async def _read_version(
        *,
        url: str,
        group: str,
        version: str,
        preferred: bool,
        context: auth.APIContext,
) -> Collection[references.Resource]:
    try:
        rsp = await errors.parse_response(await context.session.get(url))
    except errors.APINotFoundError:
        # This happens when the last and the only resource of a group/version
        # has been deleted, the whole group/version is gone, and we rescan it.
        return set()
    else:
        # Note: builtins' singulars are empty strings in K3s (reasons unknown):
        # fall back to the lowercased kind so that the selectors could match.
        return {
            references.Resource(
                group=group,
                version=version,
                kind=resource['kind'],
                plural=resource['name'],
                singular=resource['singularName'] or resource['kind'].lower(),
                shortcuts=frozenset(resource.get('shortNames', [])),
                categories=frozenset(resource.get('categories', [])),
                subresources=frozenset(
                    subresource['name'].split('/', 1)[-1]
                    for subresource in rsp['resources']
                    if subresource['name'].startswith(f'{resource["name"]}/')
                ),
                namespaced=resource['namespaced'],
                preferred=preferred,
                verbs=frozenset(resource.get('verbs', [])),
            )
            for resource in rsp['resources']
            if '/' not in resource['name']
        }



import pytest

import kopf
from kopf.reactor.admission import admission_webhook_server
from kopf.structs.primitives import Container


async def webhookfn(*_, **__):
    pass


async def test_requires_webserver_if_webhooks_are_defined(
        settings, registry, insights, resource, k8s_mocked):

    @kopf.on.validate(*resource, registry=registry)
    def fn_v(**_): ...

    @kopf.on.mutate(*resource, registry=registry)
    def fn_m(**_): ...

    container = Container()
    with pytest.raises(Exception) as err:
        settings.admission.server = None
        await admission_webhook_server(
            registry=registry,
            settings=settings,
            insights=insights,
            container=container,
            webhookfn=webhookfn,
        )

    assert "Admission handlers exist, but no admission server/tunnel" in str(err.value)


async def test_configures_client_configs(
        settings, registry, insights, resource, k8s_mocked):

    async def server(_):
        yield {'url': 'https://hostname/'}

    container = Container()
    settings.admission.server = server
    await admission_webhook_server(
        registry=registry,
        settings=settings,
        insights=insights,
        container=container,
        webhookfn=webhookfn,
    )

    assert container.get_nowait() == {'url': 'https://hostname/'}



import logging

import aiohttp.web
import pytest

from kopf.engines.loggers import LocalObjectLogger
from kopf.reactor.effects import patch_and_check
from kopf.structs.bodies import Body
from kopf.structs.patches import Patch

# Assume that the underlying patch_obj() is already tested with/without status as a sub-resource.
# Assume that the underlying diff() is already tested with left/right/full scopes and all values.
# Test ONLY the logging/warning on patch-vs-response inconsistencies here.


@pytest.mark.parametrize('patch, response', [

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 id='response-exact'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y'}, 'status': {'s': 't'}, 'metadata': '...'},
                 id='response-root-extra'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y', 'extra': '...'}, 'status': {'s': 't'}},
                 id='response-spec-extra'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y'}, 'status': {'s': 't', 'extra': '...'}},
                 id='response-status-extra'),

    pytest.param({'spec': {'x': None}, 'status': {'s': None}},
                 {'spec': {}, 'status': {}},
                 id='response-clean'),

    # False-positive inconsistencies for K8s-managed fields.
    pytest.param({'metadata': {'annotations': {}}}, {'metadata': {}}, id='false-annotations'),
    pytest.param({'metadata': {'finalizers': []}}, {'metadata': {}}, id='false-finalizers'),
    pytest.param({'metadata': {'labels': {}}}, {'metadata': {}}, id='false-labels'),

])
async def test_patching_without_inconsistencies(
        resource, namespace, settings, caplog, assert_logs, version_api,
        aresponses, hostname, resp_mocker,
        patch, response):
    caplog.set_level(logging.DEBUG)

    url = resource.get_url(namespace=namespace, name='name1')
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response(response))
    aresponses.add(hostname, url, 'patch', patch_mock)

    body = Body({'metadata': {'namespace': namespace, 'name': 'name1'}})
    logger = LocalObjectLogger(body=body, settings=settings)
    await patch_and_check(
        resource=resource,
        body=body,
        patch=Patch(patch),
        logger=logger,
    )

    assert_logs([
        "Patching with:",
    ], prohibited=[
        "Patching failed with inconsistencies:",
    ])


@pytest.mark.parametrize('patch, response', [

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {},
                 id='response-empty'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y'}},
                 id='response-status-lost'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'status': {'s': 't'}},
                 id='response-spec-lost'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'not-y'}, 'status': {'s': 't'}},
                 id='response-spec-altered'),

    pytest.param({'spec': {'x': 'y'}, 'status': {'s': 't'}},
                 {'spec': {'x': 'y'}, 'status': {'s': 'not-t'}},
                 id='response-status-altered'),

    pytest.param({'spec': {'x': None}, 'status': {'s': None}},
                 {'spec': {'x': 'y'}, 'status': {}},
                 id='response-spec-undeleted'),

    pytest.param({'spec': {'x': None}, 'status': {'s': None}},
                 {'spec': {}, 'status': {'s': 't'}},
                 id='response-status-undeleted'),

    # True-positive inconsistencies for K8s-managed fields with possible false-positives.
    pytest.param({'metadata': {'annotations': {'x': 'y'}}}, {'metadata': {}}, id='true-annotations'),
    pytest.param({'metadata': {'finalizers': ['x', 'y']}}, {'metadata': {}}, id='true-finalizers'),
    pytest.param({'metadata': {'labels': {'x': 'y'}}}, {'metadata': {}}, id='true-labels'),

])
async def test_patching_with_inconsistencies(
        resource, namespace, settings, caplog, assert_logs, version_api,
        aresponses, hostname, resp_mocker,
        patch, response):
    caplog.set_level(logging.DEBUG)

    url = resource.get_url(namespace=namespace, name='name1')
    patch_mock = resp_mocker(return_value=aiohttp.web.json_response(response))
    aresponses.add(hostname, url, 'patch', patch_mock)

    body = Body({'metadata': {'namespace': namespace, 'name': 'name1'}})
    logger = LocalObjectLogger(body=body, settings=settings)
    await patch_and_check(
        resource=resource,
        body=body,
        patch=Patch(patch),
        logger=logger,
    )

    assert_logs([
        "Patching with:",
        "Patching failed with inconsistencies:",
    ])


async def test_patching_with_disappearance(
        resource, namespace, settings, caplog, assert_logs, version_api,
        aresponses, hostname, resp_mocker):
    caplog.set_level(logging.DEBUG)

    patch = {'spec': {'x': 'y'}, 'status': {'s': 't'}}  # irrelevant
    url = resource.get_url(namespace=namespace, name='name1')
    patch_mock = resp_mocker(return_value=aresponses.Response(status=404))
    aresponses.add(hostname, url, 'patch', patch_mock)

    body = Body({'metadata': {'namespace': namespace, 'name': 'name1'}})
    logger = LocalObjectLogger(body=body, settings=settings)
    await patch_and_check(
        resource=resource,
        body=body,
        patch=Patch(patch),
        logger=logger,
    )

    assert_logs([
        "Patching with:",
        "Patching was skipped: the object does not exist anymore",
    ], prohibited=[
        "inconsistencies"
    ])



import asyncio
import base64
import copy
import json
import logging
import re
import urllib.parse
from typing import Any, Collection, Dict, Iterable, List, Mapping, Optional

from typing_extensions import Literal, TypedDict

from kopf.clients import creating, errors, patching
from kopf.engines import loggers
from kopf.reactor import causation, handling, lifecycles, registries
from kopf.storage import states
from kopf.structs import bodies, configuration, containers, ephemera, filters, \
                         handlers, ids, patches, primitives, references, reviews

logger = logging.getLogger(__name__)


class AdmissionError(handling.PermanentError):
    """
    Raised by admission handlers when an API operation under check is bad.

    An admission error behaves the same as `kopf.PermanentError`, but provides
    admission-specific payload for the response: a message & a numeric code.

    This error type is preferred when selecting only one error to report back
    to apiservers as the admission review result -- in case multiple handlers
    are called in one admission request, i.e. when the webhook endpoints
    are not mapped to the handler ids (e.g. when configured manually).
    """
    def __init__(
            self,
            message: Optional[str] = '',
            code: Optional[int] = 500,
    ) -> None:
        super().__init__(message)
        self.code = code


class WebhookError(Exception):
    """
    Raised when a webhook request is bad, not an API operation under check.
    """


class MissingDataError(WebhookError):
    """ An admission is requested but some expected data are missing. """


class UnknownResourceError(WebhookError):
    """ An admission is made for a resource that the operator does not have. """


class AmbiguousResourceError(WebhookError):
    """ An admission is made for one resource, but we (somehow) found a few. """


async def serve_admission_request(
        # Required for all webhook servers, meaningless without it:
        request: reviews.Request,
        *,
        # Optional for webhook servers that can recognise this information:
        headers: Optional[Mapping[str, str]] = None,
        sslpeer: Optional[Mapping[str, Any]] = None,
        webhook: Optional[ids.HandlerId] = None,
        reason: Optional[handlers.WebhookType] = None,  # TODO: undocumented: requires typing clarity!
        # Injected by partial() from spawn_tasks():
        settings: configuration.OperatorSettings,
        memories: containers.ResourceMemories,
        memobase: ephemera.AnyMemo,
        registry: registries.OperatorRegistry,
        insights: references.Insights,
        indices: ephemera.Indices,
) -> reviews.Response:
    """
    The actual and the only implementation of the `WebhookFn` protocol.

    This function is passed to all webhook servers/tunnels to be called
    whenever a new admission request is received.

    Some parameters are provided by the framework itself via partial binding,
    so that the resulting function matches the `WebhookFn` protocol. Other
    parameters are passed by the webhook servers when they call the function.
    """

    # Reconstruct the cause specially for web handlers.
    resource = find_resource(request=request, insights=insights)
    operation = request.get('request', {}).get('operation')
    userinfo = request.get('request', {}).get('userInfo')
    new_body = request.get('request', {}).get('object')
    old_body = request.get('request', {}).get('oldObject')
    raw_body = new_body if new_body is not None else old_body
    if userinfo is None:
        raise MissingDataError("User info is missing from the admission request.")
    if raw_body is None:
        raise MissingDataError("Either old or new object is missing from the admission request.")

    memory = await memories.recall(raw_body, memo=memobase, ephemeral=operation=='CREATE')
    body = bodies.Body(raw_body)
    patch = patches.Patch()
    warnings: List[str] = []
    cause = causation.ResourceWebhookCause(
        resource=resource,
        indices=indices,
        logger=loggers.LocalObjectLogger(body=body, settings=settings),
        patch=patch,
        memo=memory.memo,
        body=body,
        userinfo=userinfo,
        warnings=warnings,
        operation=operation,
        dryrun=bool(request.get('request', {}).get('dryRun')),
        sslpeer=sslpeer if sslpeer is not None else {},  # ensure a mapping even if not provided.
        headers=headers if headers is not None else {},  # ensure a mapping even if not provided.
        webhook=webhook,
        reason=reason,
    )

    # Retrieve the handlers to be executed; maybe only one if the webhook server provides a hint.
    handlers_ = registry._resource_webhooks.get_handlers(cause)
    state = states.State.from_scratch().with_handlers(handlers_)
    outcomes = await handling.execute_handlers_once(
        lifecycle=lifecycles.all_at_once,
        settings=settings,
        handlers=handlers_,
        cause=cause,
        state=state,
        default_errors=handlers.ErrorsMode.PERMANENT,
    )

    # Construct the response as per Kubernetes's conventions and expectations.
    response = build_response(
        request=request,
        outcomes=outcomes,
        warnings=warnings,
        jsonpatch=patch.as_json_patch(),
    )
    return response


def find_resource(
        *,
        request: reviews.Request,
        insights: references.Insights,
) -> references.Resource:
    """
    Identify the requested resource by its meta-information (as discovered).
    """
    # NB: Absent keys in the request are not acceptable, they must be provided.
    request_payload: reviews.RequestPayload = request['request']
    request_resource: reviews.RequestResource = request_payload['resource']
    group = request_resource['group']
    version = request_resource['version']
    plural = request_resource['resource']
    selector = references.Selector(group=group, version=version, plural=plural)
    resources = selector.select(insights.resources)
    if not resources:
        raise UnknownResourceError(f"The specified resource has no handlers: {request_resource}")
    elif len(resources) > 1:
        raise AmbiguousResourceError(f"The specified resource is ambiguous: {request_resource}")
    else:
        return list(resources)[0]


def build_response(
        *,
        request: reviews.Request,
        outcomes: Mapping[ids.HandlerId, states.HandlerOutcome],
        warnings: Collection[str],
        jsonpatch: patches.JSONPatch,
) -> reviews.Response:
    """
    Construct the admission review response to a review request.
    """
    allowed = all(outcome.exception is None for id, outcome in outcomes.items())
    response = reviews.Response(
        apiVersion=request.get('apiVersion', 'admission.k8s.io/v1'),
        kind=request.get('kind', 'AdmissionReview'),
        response=reviews.ResponsePayload(
            uid=request.get('request', {}).get('uid', ''),
            allowed=allowed))
    if warnings:
        response['response']['warnings'] = [str(warning) for warning in warnings]
    if jsonpatch:
        encoded_patch: str = base64.b64encode(json.dumps(jsonpatch).encode('utf-8')).decode('ascii')
        response['response']['patch'] = encoded_patch
        response['response']['patchType'] = 'JSONPatch'

    # Prefer specialised admission errors to all other errors, Kopf's own errors to arbitrary ones.
    errors = [outcome.exception for outcome in outcomes.values() if outcome.exception is not None]
    errors.sort(key=lambda error: (
        0 if isinstance(error, AdmissionError) else
        1 if isinstance(error, handling.PermanentError) else
        2 if isinstance(error, handling.TemporaryError) else
        9
    ))
    if errors:
        response['response']['status'] = reviews.ResponseStatus(
            message=str(errors[0]) or repr(errors[0]),
            code=(errors[0].code if isinstance(errors[0], AdmissionError) else None) or 500,
        )
    return response


async def admission_webhook_server(
        *,
        settings: configuration.OperatorSettings,
        registry: registries.OperatorRegistry,
        insights: references.Insights,
        webhookfn: reviews.WebhookFn,
        container: primitives.Container[reviews.WebhookClientConfig],
) -> None:

    # Verify that the operator is configured properly (after the startup activities are done).
    has_admission = bool(registry._resource_webhooks.get_all_handlers())
    if settings.admission.server is None and has_admission:
        raise Exception(
            "Admission handlers exist, but no admission server/tunnel is configured "
            "in `settings.admission.server`. "
            "More: https://kopf.readthedocs.io/en/stable/admission/")

    # Do not start the endpoints until resources are scanned.
    # Otherwise, we generate 404 "Not Found" for requests that arrive too early.
    await insights.ready_resources.wait()

    # Communicate all the client configs the server yields: both the initial one and the updates.
    # On each such change, the configuration manager will wake up and reconfigure the webhooks.
    if settings.admission.server is not None:
        async for client_config in settings.admission.server(webhookfn):
            await container.set(client_config)
    else:
        await asyncio.Event().wait()


async def validating_configuration_manager(
        *,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        insights: references.Insights,
        container: primitives.Container[reviews.WebhookClientConfig],
) -> None:
    await configuration_manager(
        reason=handlers.WebhookType.VALIDATING,
        selector=references.VALIDATING_WEBHOOK,
        registry=registry, settings=settings,
        insights=insights, container=container,
    )


async def mutating_configuration_manager(
        *,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        insights: references.Insights,
        container: primitives.Container[reviews.WebhookClientConfig],
) -> None:
    await configuration_manager(
        reason=handlers.WebhookType.MUTATING,
        selector=references.MUTATING_WEBHOOK,
        registry=registry, settings=settings,
        insights=insights, container=container,
    )


async def configuration_manager(
        *,
        reason: handlers.WebhookType,
        selector: references.Selector,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        insights: references.Insights,
        container: primitives.Container[reviews.WebhookClientConfig],
) -> None:
    """
    Manage the webhook configurations dynamically.

    This is one of an operator's root tasks that run forever.
    If exited, the whole operator exits as by an error.

    The manager waits for changes in one of these:

    * Observed resources in the cluster (via insights).
    * A new webhook client config yielded by the webhook server.

    On either of these occasion, the manager rebuilds the webhook configuration
    and applies it to the specified configuration resources in the cluster
    (for which it needs some RBAC permissions).
    Besides, it also creates an webhook configuration resource if it is absent.
    """

    # Do nothing if not managed. The root task cannot be skipped from creation,
    # since the managed mode is only set at the startup activities.
    if settings.admission.managed is None:
        await asyncio.Event().wait()
        return

    # Wait until the prerequisites for managing are available (scanned from the cluster).
    await insights.ready_resources.wait()
    resource = await insights.backbone.wait_for(selector)
    all_handlers = registry._resource_webhooks.get_all_handlers()
    all_handlers = [h for h in all_handlers if h.reason == reason]

    # Optionally (if configured), pre-create the configuration objects if they are absent.
    # Use the try-or-fail strategy instead of check-and-do -- to reduce the RBAC requirements.
    try:
        await creating.create_obj(resource=resource, name=settings.admission.managed)
    except errors.APIConflictError:
        pass  # exists already
    except errors.APIForbiddenError:
        logger.error(f"Not enough RBAC permissions to create a {resource}.")
        raise

    # Execute either when actually changed (yielded from the webhook server),
    # or the condition is chain-notified (from the insights: on resources/namespaces revision).
    # Ignore inconsistencies: they are expected -- the server fills the defaults.
    client_config: Optional[reviews.WebhookClientConfig] = None
    try:
        async for client_config in container.as_changed():
            logger.info(f"Reconfiguring the {reason.value} webhook {settings.admission.managed}.")
            webhooks = build_webhooks(
                all_handlers,
                resources=insights.resources,
                name_suffix=settings.admission.managed,
                client_config=client_config)
            await patching.patch_obj(
                resource=resource,
                namespace=None,
                name=settings.admission.managed,
                patch=patches.Patch({'webhooks': webhooks}),
            )
    finally:
        # Attempt to remove all managed webhooks, except for the strict ones.
        if client_config is not None:
            logger.info(f"Cleaning up the admission webhook {settings.admission.managed}.")
            webhooks = build_webhooks(
                all_handlers,
                resources=insights.resources,
                name_suffix=settings.admission.managed,
                client_config=client_config,
                persistent_only=True)
            await patching.patch_obj(
                resource=resource,
                namespace=None,
                name=settings.admission.managed,
                patch=patches.Patch({'webhooks': webhooks}),
            )


def build_webhooks(
        handlers_: Iterable[handlers.ResourceWebhookHandler],
        *,
        resources: Iterable[references.Resource],
        name_suffix: str,
        client_config: reviews.WebhookClientConfig,
        persistent_only: bool = False,
) -> List[Dict[str, Any]]:
    """
    Construct the content for ``[Validating|Mutating]WebhookConfiguration``.

    This function concentrates all conventions how Kopf manages the webhook.
    """
    return [
        {
            'name': _normalize_name(handler.id, suffix=name_suffix),
            'sideEffects': 'NoneOnDryRun' if handler.side_effects else 'None',
            'failurePolicy': 'Ignore' if handler.ignore_failures else 'Fail',
            'matchPolicy': 'Equivalent',
            'rules': [
                {
                    'apiGroups': [resource.group],
                    'apiVersions': [resource.version],
                    'resources': [resource.plural],
                    'operations': ['*'] if handler.operation is None else [handler.operation],
                    'scope': '*',  # doesn't matter since a specific resource is used.
                }
                for resource in resources
                if handler.selector is not None  # None is used only in sub-handlers, ignore here.
                if handler.selector.check(resource)
            ],
            'objectSelector': _build_labels_selector(handler.labels),
            'clientConfig': _inject_handler_id(client_config, handler.id),
            'timeoutSeconds': 30,  # a permitted maximum is 30.
            'admissionReviewVersions': ['v1', 'v1beta1'],  # only those understood by Kopf itself.
        }
        for handler in handlers_
        if not persistent_only or handler.persistent
    ]


class MatchExpression(TypedDict, total=False):
    key: str
    operator: Literal['Exists', 'DoesNotExist', 'In', 'NotIn']
    values: Optional[Collection[str]]


def _build_labels_selector(labels: Optional[filters.MetaFilter]) -> Optional[Mapping[str, Any]]:
    # https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#resources-that-support-set-based-requirements
    exprs: Collection[MatchExpression] = [
        {'key': key, 'operator': 'Exists'} if val is filters.MetaFilterToken.PRESENT else
        {'key': key, 'operator': 'DoesNotExist'} if val is filters.MetaFilterToken.ABSENT else
        {'key': key, 'operator': 'In', 'values': [str(val)]}
        for key, val in (labels or {}).items()
        if not callable(val)
    ]
    return {'matchExpressions': exprs} if exprs else None


BAD_WEBHOOK_NAME = re.compile(r'[^\w\d\.-]')


def _normalize_name(id: ids.HandlerId, suffix: str) -> str:
    """
    Normalize the webhook name to what Kubernetes accepts as normal.

    The restriction is: *a lowercase RFC 1123 subdomain must consist
    of lower case alphanumeric characters, \'-\' or \'.\',
    and must start and end with an alphanumeric character.*

    The actual name is not that important, it is for informational purposes
    only. In the managed configurations, it will be rewritten every time.
    """
    name = f'{id}'.replace('/', '.').replace('_', '-')  # common cases, for beauty
    name = BAD_WEBHOOK_NAME.sub(lambda s: s.group(0).encode('utf-8').hex(), name)  # uncommon cases
    return f'{name}.{suffix}' if suffix else name


def _inject_handler_id(config: reviews.WebhookClientConfig, id: ids.HandlerId) -> reviews.WebhookClientConfig:
    config = copy.deepcopy(config)

    url_id = urllib.parse.quote(id)
    url = config.get('url')
    if url is not None:
        config['url'] = f'{url.rstrip("/")}/{url_id}'

    service = config.get('service')
    if service is not None:
        path = service.get('path', '')
        service['path'] = f"{path}/{url_id}"

    return config



import asyncio

import async_timeout
import pytest

from kopf.structs.primitives import condition_chain


async def test_no_triggering():
    source = asyncio.Condition()
    target = asyncio.Condition()
    task = asyncio.create_task(condition_chain(source, target))
    try:

        with pytest.raises(asyncio.TimeoutError):
            with async_timeout.timeout(0.1) as timeout:
                async with target:
                    await target.wait()

        assert timeout.expired

    finally:
        task.cancel()
        await asyncio.wait([task])


async def test_triggering(event_loop, timer):
    source = asyncio.Condition()
    target = asyncio.Condition()
    task = asyncio.create_task(condition_chain(source, target))
    try:

        async def delayed_trigger():
            async with source:
                source.notify_all()

        event_loop.call_later(0.1, asyncio.create_task, delayed_trigger())

        with timer, async_timeout.timeout(10) as timeout:
            async with target:
                await target.wait()

        assert not timeout.expired
        assert 0.1 <= timer.seconds <= 0.2

    finally:
        task.cancel()
        await asyncio.wait([task])



import base64
import json
import ssl

import aiohttp
import pytest

from kopf.reactor.admission import AmbiguousResourceError, MissingDataError, \
                                   UnknownResourceError, WebhookError
from kopf.toolkits.webhooks import WebhookK3dServer, WebhookMinikubeServer, WebhookServer


async def test_starts_as_http_ipv4(responder):
    server = WebhookServer(addr='127.0.0.1', port=22533, path='/p1/p2', insecure=True)
    async for client_config in server(responder.fn):
        assert client_config['url'] == 'http://127.0.0.1:22533/p1/p2'
        assert 'caBundle' not in client_config
        break  # do not sleep


async def test_starts_as_http_ipv6(responder):
    server = WebhookServer(addr='::1', port=22533, path='/p1/p2', insecure=True)
    async for client_config in server(responder.fn):
        assert client_config['url'] == 'http://[::1]:22533/p1/p2'
        assert 'caBundle' not in client_config
        break  # do not sleep


async def test_unspecified_port_allocates_a_random_port(responder):
    server1 = WebhookServer(addr='127.0.0.1', path='/p1/p2', insecure=True)
    server2 = WebhookServer(addr='127.0.0.1', path='/p1/p2', insecure=True)
    async for client_config1 in server1(responder.fn):
        async for client_config2 in server2(responder.fn):
            assert client_config1['url'] != client_config2['url']
            break  # do not sleep
        break  # do not sleep


async def test_unspecified_addr_uses_all_interfaces(responder, caplog, assert_logs):
    caplog.set_level(0)
    server = WebhookServer(port=22533, path='/p1/p2', insecure=True)
    async for client_config in server(responder.fn):
        assert client_config['url'] == 'http://localhost:22533/p1/p2'
        break  # do not sleep
    assert_logs([r"Listening for webhooks at http://\*:22533/p1/p2"])


async def test_webhookserver_starts_as_https_with_selfsigned_cert(
        responder):
    server = WebhookServer(addr='127.0.0.1', port=22533, path='/p1/p2', host='somehost')
    async for client_config in server(responder.fn):
        assert client_config['url'] == 'https://somehost:22533/p1/p2'
        assert 'caBundle' in client_config  # regardless of the value
        break  # do not sleep


async def test_webhookserver_starts_as_https_with_provided_cert(
        certfile, pkeyfile, certpkey, responder):
    server = WebhookServer(port=22533, certfile=certfile, pkeyfile=pkeyfile)
    async for client_config in server(responder.fn):
        assert client_config['url'] == 'https://localhost:22533'
        assert base64.b64decode(client_config['caBundle']) == certpkey[0]
        break  # do not sleep


@pytest.mark.parametrize('cls, url', [
    (WebhookK3dServer, 'https://host.k3d.internal:22533/p1/p2'),
    (WebhookMinikubeServer, 'https://host.minikube.internal:22533/p1/p2'),
])
async def test_webhookserver_flavours_inject_hostnames(
        certfile, pkeyfile, certpkey, responder, cls, url):
    server = cls(port=22533, certfile=certfile, pkeyfile=pkeyfile, path='/p1/p2')
    async for client_config in server(responder.fn):
        assert client_config['url'] == url
        break  # do not sleep


@pytest.mark.usefixtures('no_sslproto_warnings')
async def test_webhookserver_serves(
        certfile, pkeyfile, responder, adm_request):
    responder.fut.set_result({'hello': 'world'})
    server = WebhookServer(certfile=certfile, pkeyfile=pkeyfile)
    async for client_config in server(responder.fn):
        cadata = base64.b64decode(client_config['caBundle']).decode('ascii')
        sslctx = ssl.create_default_context(cadata=cadata)
        async with aiohttp.ClientSession() as client:
            async with client.post(client_config['url'], ssl=sslctx, json=adm_request) as resp:
                text = await resp.text()
                assert text == '{"hello": "world"}'
                assert resp.status == 200
        break  # do not sleep


@pytest.mark.parametrize('code, error', [
    (500, Exception),
    (400, WebhookError),
    (400, WebhookError),
    (400, MissingDataError),
    (404, UnknownResourceError),
    (409, AmbiguousResourceError),
    (400, lambda: json.JSONDecodeError('...', '...', 0)),
])
@pytest.mark.usefixtures('no_sslproto_warnings')
async def test_webhookserver_errors(
        certfile, pkeyfile, responder, adm_request, code, error):
    responder.fut.set_exception(error())
    server = WebhookServer(certfile=certfile, pkeyfile=pkeyfile)
    async for client_config in server(responder.fn):
        cadata = base64.b64decode(client_config['caBundle']).decode('ascii')
        sslctx = ssl.create_default_context(cadata=cadata)
        async with aiohttp.ClientSession() as client:
            async with client.post(client_config['url'], ssl=sslctx, json=adm_request) as resp:
                assert resp.status == code
        break  # do not sleep



"""
Callback signatures for typing.

Since these signatures contain a lot of copy-pasted kwargs and are
not so important for the codebase, they are moved to this separate module.
"""
import datetime
import logging
from typing import TYPE_CHECKING, Any, Callable, Collection, \
                   Coroutine, List, NewType, Optional, TypeVar, Union

from kopf.structs import bodies, configuration, diffs, ephemera, \
                         patches, primitives, references, reviews

# A specialised type to highlight the purpose or origin of the data of type Any,
# to not be mixed with other arbitrary Any values, where it is indeed "any".
Result = NewType('Result', object)

# An internal typing hack to show that it can be a sync fn with the result,
# or an async fn which returns a coroutine which returns the result.
# Used in sync-and-async protocols only, and never exposed to other modules.
_SyncOrAsyncResult = Union[Optional[Result], Coroutine[None, None, Optional[Result]]]

# A generic sync-or-async callable with no args/kwargs checks (unlike in protocols).
# Used for the BaseHandler and generic invocation methods (which do not care about protocols).
BaseFn = Callable[..., _SyncOrAsyncResult]

LoggerType = Union[logging.Logger, logging.LoggerAdapter]

if not TYPE_CHECKING:  # pragma: nocover
    # Define unspecified protocols for the runtime annotations -- to avoid "quoting".
    ActivityFn = Callable[..., _SyncOrAsyncResult]
    ResourceIndexingFn = Callable[..., _SyncOrAsyncResult]
    ResourceWatchingFn = Callable[..., _SyncOrAsyncResult]
    ResourceChangingFn = Callable[..., _SyncOrAsyncResult]
    ResourceWebhookFn = Callable[..., None]
    ResourceDaemonFn = Callable[..., _SyncOrAsyncResult]
    ResourceTimerFn = Callable[..., _SyncOrAsyncResult]
    WhenFilterFn = Callable[..., bool]
    MetaFilterFn = Callable[..., bool]
else:
    from mypy_extensions import Arg, DefaultNamedArg, KwArg, NamedArg, VarArg

    # TODO: Try using ParamSpec to support index type checking in callbacks 
    # when PEP 612 is released (https://www.python.org/dev/peps/pep-0612/)
    ActivityFn = Callable[
        [
            NamedArg(configuration.OperatorSettings, "settings"),
            NamedArg(ephemera.Index, "*"),
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    ResourceIndexingFn = Callable[
        [
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    ResourceWatchingFn = Callable[
        [
            NamedArg(str, "type"),
            NamedArg(bodies.RawEvent, "event"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    ResourceChangingFn = Callable[
        [
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(str, "reason"),
            NamedArg(diffs.Diff, "diff"),
            NamedArg(Optional[Union[bodies.BodyEssence, Any]], "old"),
            NamedArg(Optional[Union[bodies.BodyEssence, Any]], "new"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    ResourceWebhookFn = Callable[
        [
            NamedArg(bool, "dryrun"),
            NamedArg(List[str], "warnings"),  # mutable!
            NamedArg(reviews.UserInfo, "userinfo"),
            NamedArg(reviews.SSLPeer, "sslpeer"),
            NamedArg(reviews.Headers, "headers"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        None
    ]

    ResourceDaemonFn = Callable[
        [
            NamedArg(primitives.SyncAsyncDaemonStopperChecker, "stopped"),
            NamedArg(int, "retry"),
            NamedArg(datetime.datetime, "started"),
            NamedArg(datetime.timedelta, "runtime"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    ResourceTimerFn = Callable[
        [
            NamedArg(ephemera.Index, "*"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        _SyncOrAsyncResult
    ]

    WhenFilterFn = Callable[
        [
            NamedArg(str, "type"),
            NamedArg(bodies.RawEvent, "event"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(diffs.Diff, "diff"),
            NamedArg(Optional[Union[bodies.BodyEssence, Any]], "old"),
            NamedArg(Optional[Union[bodies.BodyEssence, Any]], "new"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        bool
    ]

    MetaFilterFn = Callable[
        [
            Arg(Any, "value"),
            NamedArg(str, "type"),
            NamedArg(bodies.Annotations, "annotations"),
            NamedArg(bodies.Labels, "labels"),
            NamedArg(bodies.Body, "body"),
            NamedArg(bodies.Meta, "meta"),
            NamedArg(bodies.Spec, "spec"),
            NamedArg(bodies.Status, "status"),
            NamedArg(references.Resource, "resource"),
            NamedArg(Optional[str], "uid"),
            NamedArg(Optional[str], "name"),
            NamedArg(Optional[str], "namespace"),
            NamedArg(patches.Patch, "patch"),
            NamedArg(LoggerType, "logger"),
            NamedArg(ephemera.AnyMemo, "memo"),
            DefaultNamedArg(Any, "param"),
            KwArg(Any),
        ],
        bool
    ]

ResourceSpawningFn = Union[ResourceDaemonFn, ResourceTimerFn]
_FnT = TypeVar('_FnT', WhenFilterFn, MetaFilterFn)


def not_(fn: _FnT) -> _FnT:
    def not_fn(*args: Any, **kwargs: Any) -> bool:
        return not fn(*args, **kwargs)
    return not_fn


def all_(fns: Collection[_FnT]) -> _FnT:
    def all_fn(*args: Any, **kwargs: Any) -> bool:
        return all(fn(*args, **kwargs) for fn in fns)
    return all_fn


def any_(fns: Collection[_FnT]) -> _FnT:
    def any_fn(*args: Any, **kwargs: Any) -> bool:
        return any(fn(*args, **kwargs) for fn in fns)
    return any_fn


def none_(fns: Collection[_FnT]) -> _FnT:
    def none_fn(*args: Any, **kwargs: Any) -> bool:
        return not any(fn(*args, **kwargs) for fn in fns)
    return none_fn
